#include <iostream>
#include <fstream>
#include <string>
#include <windows.h>
using namespace std;

void print_usage(const char* prog)
{
  cout << "Usage:\n" << prog << " <path to private headers> "
       << "<path to put public headers into>\n"
       << "The two paths must not be the same\n";
}

/// <summary>
/// Converts error codes from GetLastError to their string from
/// </summary>
/// <remarks>
/// The return value should be freed with LocalFree when it is no longer needed
/// </remarks>
/// <param name="err_code">
/// error code to convert
/// </param>
/// <returns>
/// string form of the error code
/// NULL if unable to convert
/// </returns>
const char* StrError(DWORD err_code)
{
  LPSTR back = NULL;
  
  FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,err_code,0,(LPSTR)&back,0,NULL);
  
  return back;
}

/// <summary>
/// Converts a private header file to a public one
/// </summary>
/// <remarks>
/// src must not be the same path as dst.  It is up to the caller to ensure
/// this
/// </remarks>
/// <param name="src">
/// path to the private header to convert
/// </param>
/// <param name="dst">
/// path to where to put the public header
/// </param>
void process_file(const string& src,const string& dst)
{
  ifstream in(src.c_str());
  ofstream out(dst.c_str());
  
  string line;
  bool output = true;
  bool next_output = false;
  while (getline(in,line,'\n'))
  {
    // check if output should be turned on at the start of the next line (which
    // the code is now at)
    if (!output && next_output)
    {
      output = true;
    }
    
    // check if the line should be output
    if (line.find("PUBLIC_HEADER") != string::npos)
    {
      // limit turning on/off output to starting at the beginning of a line so
      // it can be commented out if needed
      if (line.find("#ifndef") == 0)
      {
        output      = false;
        next_output = false;
      }
      else if (line.find("#endif") == 0)
      {
        next_output = true;
      }
    }
    
    // output the line if it should be
    if (output)
    {
      out << line << '\n';
    }
  }
  in.close();
  
  // if the output file is essentially empty, delete it
  if (out.tellp() <= 2)
  {
    out.close();
    
    if (!DeleteFile(dst.c_str()))
    {
      const char* err_code = StrError(GetLastError());
      cerr << "Failed to delete empty file \"" << dst << "\", reason: "
           << err_code << '\n';
      LocalFree((HLOCAL)err_code);
    }
  }
  else
  {
    out.close();
  }
}

/// <summary>
/// Recursively searchs through src to find .h and .hpp files to convert to
/// their public header form
/// </summary>
/// <remarks>
/// src must not be the same path as dst.  It is up to the caller to ensure
/// this
/// </remarks>
/// <param name="src">
/// path to the private header directory to search through
/// </param>
/// <param name="dst">
/// path to the directory to put the public headers into
/// </param>
void search_dir(const string& src,const string& dst)
{
  // start searching the private header directory
  WIN32_FIND_DATA find_data;
  HANDLE find_handle = FindFirstFile((src + "\\*").c_str(),
    &find_data);
  if (find_handle == INVALID_HANDLE_VALUE)
  {
    const char* err_code = StrError(GetLastError());
    cerr << "Failed to open (skipping) private header directory \"" << src
      << "\", reason: " << err_code << '\n';
    LocalFree((HLOCAL)err_code);
  }
  do
  {
    if (find_data.cFileName[0] != '.')
    {
      if (find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      {
        // got a subdirectory
        
        // as a lazy workaround for CreateDirectory only able to do 1 layer at
        // a time, just create the subdirectory for public headers here.  This
        // can result in extra empty directories if the subdirectory doesn't
        // contain any headers
        string next_dst = dst + '\\' + find_data.cFileName;
        BOOL process = CreateDirectory(next_dst.c_str(),NULL);
        if (!process)
        {
          DWORD err = GetLastError();
          if (err == ERROR_ALREADY_EXISTS)
          {
            // ignore directory already exists errors
            process = true;
          }
          else
          {
            const char* err_code = StrError(err);
            cerr << "Failed to create directory \"" << next_dst << "\", "
                 << "skipping processing of \""
                 << (src + '\\' + find_data.cFileName) << "\", reason: "
                 << err_code << '\n';
            LocalFree((HLOCAL)err_code);
          }
        }
        
        if (process)
        {
          // search/process the subdirectory
          search_dir(src + '\\' + find_data.cFileName,next_dst);
        }
      }
      else if ((find_data.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE)
        || (find_data.dwFileAttributes & FILE_ATTRIBUTE_NORMAL))
      {
        // got a file, if it's a .h or .hpp file, then process it
        string filename = find_data.cFileName;
        string::size_type pos = filename.find_last_of('.');
        if (pos != string::npos)
        {
          string ext = filename.substr(pos);
          if (ext == ".h" || ext == ".hpp")
          {
            process_file(src + '\\' + filename,dst + '\\' + filename);
          }
        }
      }
    }
  } while (FindNextFile(find_handle,&find_data));
  
  FindClose(find_handle);
}

int main(int argc,const char* argv[])
{
  if (argc != 3)
  {
    print_usage(argv[0]);
    return 1;
  }
  
  if (!strcmp(argv[1],argv[2]))
  {
    cerr << "Error: the path to where to put the public headers cannot be the "
         << "same as the private header path\n";
    return 1;
  }
  
  search_dir(argv[1],argv[2]);
  
  return 0;
}
